/* ltiff.c derived from David Koblas's xpaint */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tiffio.h>
#include "loadimg.h"

static void nop(void)
{}

static int forceRGB;
static unsigned short planarconfig;
static unsigned short *redmap, *greenmap, *bluemap;
static TIFF *tif=NULL;
static unsigned short bitsPS;
static unsigned short samplesPP;
static unsigned short photometric;
static unsigned short orientation;

static int width;
static int cmapSize;
static int readLines;
static int isRGBA;

static int tryLoad(char *filename, int *loadType, int *xpixels,
 int *ypixels, int *xdpi, int *ydpi, unsigned char *cmap, char *info,
 char *reverse, int *numInterlace, int *coreFile, int realLoad)
{
 int maxval, ncolors;
 int i;
 uint16 resUnit=RESUNIT_INCH;
 uint16  extrasamples, *extratypes;
 float xres, yres;

 forceRGB = FALSE;
 isRGBA = FALSE;

 TIFFSetErrorHandler((TIFFErrorHandler) nop);
// TIFFSetWarningHandler((TIFFErrorHandler) nop);
 if (tif)
  TIFFClose(tif);

 if ((tif = TIFFOpen(filename, "rb")) == NULL)
  return(LOAD_FILETYPE_UNKNOWN);

 if (!TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bitsPS))
     bitsPS = 1;

 if (!TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &samplesPP))
     samplesPP = 1;

 if (!TIFFGetField(tif, TIFFTAG_PHOTOMETRIC, &photometric))
  return(LOAD_FILETYPE_UNKNOWN);

 if (TIFFIsTiled(tif))
  return(LOAD_FILETYPE_UNKNOWN);

 if (samplesPP != 1 && samplesPP != 3 && samplesPP != 4)
  return(LOAD_FILETYPE_UNKNOWN);

 if (photometric == PHOTOMETRIC_YCBCR || photometric == PHOTOMETRIC_MASK
//     || photometric == PHOTOMETRIC_CIELAB
     )
  return(LOAD_FILETYPE_UNKNOWN);
 
 /* we only support CMYK separation */
 if ((photometric == PHOTOMETRIC_SEPARATED) && (samplesPP != 4))
  return(LOAD_FILETYPE_UNKNOWN);

 if (!TIFFGetField (tif, TIFFTAG_EXTRASAMPLES, &extrasamples, &extratypes))
  extrasamples = 0;

 if (extrasamples)
  if ((extratypes[0] == EXTRASAMPLE_ASSOCALPHA) && (samplesPP == 4) && (photometric == PHOTOMETRIC_RGB))
   isRGBA = TRUE;

 TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, xpixels);
 TIFFGetField(tif, TIFFTAG_IMAGELENGTH, ypixels);
 TIFFGetFieldDefaulted(tif, TIFFTAG_PLANARCONFIG, &planarconfig);

 TIFFGetField(tif, TIFFTAG_RESOLUTIONUNIT, &resUnit);
  
 TIFFGetField(tif, TIFFTAG_XRESOLUTION, &xres);
 TIFFGetField(tif, TIFFTAG_YRESOLUTION, &yres);
 if (resUnit==RESUNIT_INCH)
  {
   *xdpi = (int) (xres*1000+0.5);
   *ydpi = (int) (yres*1000+0.5);
  }
 else
  if (resUnit==RESUNIT_CENTIMETER)
   {
    *xdpi = (int) (xres*1000/2.54+0.5);
    *ydpi = (int) (yres*1000/2.54+0.5);
   }
  else
   *xdpi = *ydpi = 0;

 if (planarconfig == PLANARCONFIG_SEPARATE && photometric != PHOTOMETRIC_RGB)
  return(LOAD_FILETYPE_UNKNOWN);

 ncolors = 1 << bitsPS;
 maxval = ncolors - 1;
 if (maxval == 1 && samplesPP == 1)
  {
   cmap[0]=cmap[1]=cmap[2]= (unsigned char) ((photometric == PHOTOMETRIC_MINISBLACK) ? 0 : 255);
   cmap[3]=cmap[4]=cmap[5]= (unsigned char) ((photometric == PHOTOMETRIC_MINISBLACK) ? 255 : 0);
   *loadType=LOAD_CMAP; /* note that BW is loaded as CMAP by this loader */
  }
 else
  switch (photometric)
   {
    case PHOTOMETRIC_MINISBLACK:
    case PHOTOMETRIC_MINISWHITE:
     if (ncolors<=256)
      for (i = 0; i<ncolors; i++)
       cmap[i*3]=cmap[i*3+1]=cmap[i*3+2] = (unsigned char) 
	((photometric==PHOTOMETRIC_MINISBLACK) ? (i*255)/maxval : ((maxval-i)*255)/maxval);
     else
      for (i = 0; i<256; i++)
       cmap[i*3]=cmap[i*3+1]=cmap[i*3+2] = (unsigned char) 
	((photometric==PHOTOMETRIC_MINISBLACK) ? (i*255)/255 : ((maxval-i)*255)/255);
     *loadType=LOAD_CMAP; /* note that BW is loaded as CMAP by this loader */
     break;
    case PHOTOMETRIC_PALETTE:
     if (!TIFFGetField(tif, TIFFTAG_COLORMAP, &redmap, &greenmap, &bluemap))
      return(LOAD_FILETYPE_UNKNOWN);
     if (ncolors > 0xffff)
      {
       forceRGB = TRUE;
       *loadType=LOAD_TRUE_COL;
       break;
      }
     else
      for (i = 0; i < ncolors; i++)
       {
        cmap[i*3]   = (unsigned char) (redmap[i]>>8);
        cmap[i*3+1] = (unsigned char) (greenmap[i]>>8);
        cmap[i*3+2] = (unsigned char) (bluemap[i]>>8);
       }
     *loadType=LOAD_CMAP;
     break;

     case PHOTOMETRIC_SEPARATED:
     case PHOTOMETRIC_RGB:
     case PHOTOMETRIC_CIELAB:
      *loadType = isRGBA ? LOAD_RGBA : LOAD_TRUE_COL;
      break;
    }
 if (!TIFFGetField(tif, TIFFTAG_ORIENTATION, &orientation))
     orientation = ORIENTATION_TOPLEFT;

 strcpy(info, "TIFF");

 width = *xpixels;
 cmapSize = ncolors;
 readLines = 0;

 switch (orientation)
  {
   default:
   case ORIENTATION_LEFTTOP:
   case ORIENTATION_TOPLEFT:
   case ORIENTATION_LEFTBOT:
   case ORIENTATION_BOTLEFT:
    *reverse = FALSE;
    break;
   case ORIENTATION_RIGHTTOP:
   case ORIENTATION_TOPRIGHT:
   case ORIENTATION_RIGHTBOT:
   case ORIENTATION_BOTRIGHT:
    *reverse = TRUE;
    break;
  }
 *numInterlace = 1; /* no interlace */
 *coreFile = TRUE;

 return(LOAD_OK);
}

/* loader reads a band */
static int loadBand(unsigned char *data, int  lines, int  bpl)
{
 int row, rowStart, rowInc;
 int isWide;
 unsigned char *ucp;
 unsigned short *usp;
 unsigned char *sline = NULL, *sptr0=NULL, *sptr1=NULL, *sptr2=NULL;
 int x, y;
 int maxval = (1 << bitsPS) - 1;
 int scansize = 0;

int cyan, yellow, magenta, black;

 if (planarconfig == PLANARCONFIG_SEPARATE)
  {
   scansize = TIFFStripSize(tif);
   sline = _TIFFmalloc(3 * scansize);
   sptr0 = sline;
   sptr1 = sptr0 + scansize;
   sptr2 = sptr1 + scansize;
  }
 else
  if ((sline = _TIFFmalloc(TIFFScanlineSize(tif))) == NULL)
   return(LOAD_NO_MEMORY);

 isWide = cmapSize > 256;

 rowStart = 0;
 rowInc = 1;
 /* we do not support fancy horizontal orientations */

 for (row = rowStart+readLines, y = 0; y < lines; y++, row += rowInc)
  {
   unsigned char *pp = sline;
   int bitsleft, sample0, sample1, sample2;

 usp = (unsigned short *) &data[bpl*y];
 ucp = (unsigned char *) &data[bpl*y];

#define NEXT3(p0, p1, p2) {			\
     if (bitsleft == 0) {			\
       p0++;					\
       p1++;					\
       p2++;					\
       bitsleft = 8;				\
     }					\
     bitsleft -= bitsPS;			\
     sample0 = (*p0 >> bitsleft) & maxval;	\
     sample1 = (*p1 >> bitsleft) & maxval;	\
     sample2 = (*p2 >> bitsleft) & maxval;	\
}

#define NEXT(pp) do {				\
     if (bitsleft == 0) {			\
       pp++;					\
       bitsleft = 8;				\
     }					\
     bitsleft -= bitsPS;			\
     sample0 = (*pp >> bitsleft) & maxval;	\
} while (0)

   if (planarconfig == PLANARCONFIG_SEPARATE)
    {
     if (TIFFReadEncodedStrip(tif, TIFFComputeStrip(tif, row, 0), sptr0, scansize) < 0)
      break;
     if (TIFFReadEncodedStrip(tif, TIFFComputeStrip(tif, row, 1), sptr1, scansize) < 0)
      break;
     if (TIFFReadEncodedStrip(tif, TIFFComputeStrip(tif, row, 2), sptr2, scansize) < 0)
      break;
    }
   else
    if (TIFFReadScanline(tif, sline, row, 0) < 0)
     break;

   bitsleft = 8;

   switch (photometric)
    {
     case PHOTOMETRIC_PALETTE:
      if (forceRGB)
       {
	for (x = 0; x < width; x++)
         {
	  NEXT(pp);
	  *ucp++ = (unsigned char) redmap[sample0];
	  *ucp++ = (unsigned char) greenmap[sample0];
	  *ucp++ = (unsigned char) bluemap[sample0];
	 }
	break;
       }
       //	fall through 
   case PHOTOMETRIC_MINISWHITE:
   case PHOTOMETRIC_MINISBLACK:
    if (bitsPS==16) /* scale down to 8 bits */
     for (x = 0; x < width; x++)
      {
       pp++;
       *ucp++ = *pp++;
      }
    else
     for (x = 0; x < width; x++)
      {
       NEXT(pp);
       if (isWide)
	*usp++ = (unsigned char) sample0;
       else
	*ucp++ = (unsigned char) sample0;
      }
    break;
   case PHOTOMETRIC_SEPARATED:
    if (planarconfig != PLANARCONFIG_SEPARATE)
      for (x = 0; x < width; x++)
       {
	NEXT(pp);
	cyan = 255 - sample0;
	NEXT(pp);
	magenta = 255 - sample0;
	NEXT(pp);
	yellow = 255 - sample0;
	NEXT(pp);
	black = 255 - sample0;

	*ucp++ = (unsigned char)((int)black*cyan/255);
	*ucp++ = (unsigned char)((int)black*magenta/255);
	*ucp++ = (unsigned char)((int)black*yellow/255);
       }
    break;

   case PHOTOMETRIC_RGB:
    if (planarconfig == PLANARCONFIG_SEPARATE)
     {
      unsigned char *p0 = sptr0, *p1 = sptr1, *p2 = sptr2;

      for (x = 0; x < width; x++)
       {
	NEXT3(p0, p1, p2);
	*ucp++ = (unsigned char) sample0;
	*ucp++ = (unsigned char) sample1;
	*ucp++ = (unsigned char) sample2;
       }
     }
    else
     {
      /* some code repetition, but we dont't want if clauses in the inner loop */
      if (samplesPP == 4)
       {
        if (isRGBA)
	 for (x = 0; x < width; x++)
	  {
	   NEXT(pp);
	   *ucp++ = (unsigned char) sample0;
	   NEXT(pp);
	   *ucp++ = (unsigned char) sample0;
	   NEXT(pp);
	   *ucp++ = (unsigned char) sample0;
	   NEXT(pp);
	   *ucp++ = (unsigned char) sample0;
	  }
	else
	 for (x = 0; x < width; x++)
	  {
	   NEXT(pp);
	   *ucp++ = (unsigned char) sample0;
	   NEXT(pp);
	   *ucp++ = (unsigned char) sample0;
	   NEXT(pp);
	   *ucp++ = (unsigned char) sample0;
	   NEXT(pp); /* discard the extra channel */
	  }
       }
      else
       for (x = 0; x < width; x++)
	{
 	 NEXT(pp);
	 *ucp++ = (unsigned char) sample0;
	 NEXT(pp);
	 *ucp++ = (unsigned char) sample0;
	 NEXT(pp);
	 *ucp++ = (unsigned char) sample0;
	}
     }
    break;

   case PHOTOMETRIC_CIELAB:
    /* preliminary loading directly into RGB, so we can check the saved values */
    for (x = 0; x < width; x++)
     {
      NEXT(pp);
      *ucp++ = (unsigned char) sample0;
      NEXT(pp);
      *ucp++ = (unsigned char) sample0;
      NEXT(pp);
      *ucp++ = (unsigned char) sample0;
     }
    break;


   case PHOTOMETRIC_MASK:
    for (x = 0; x < width; x++)
     {
      NEXT(pp);
      *ucp++ = (unsigned char) (sample0 ? 1 : 0);
     }
    break;
   }
 }

 if (sline)
  _TIFFfree(sline);

 readLines += lines;
 return(LOAD_OK);
}

/* closes the input file and releases any temporary storage */
static int endLoad()
{
 if (tif)
  TIFFClose(tif);
 tif = NULL;
 return(LOAD_OK);
}

/* register TIFF image loader */
void registerLoadTIFF()
{
 registerLoader(*tryLoad, *loadBand, *endLoad);
}
