#define REGISTER
/* register.c */

#include <sys/types.h>
#include <unistd.h>
#ifndef __MINGW32__
#include <sys/wait.h>
#endif
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <ctype.h>
#include "print.h"

/* these functions are available to the calling program */
getPrinterInfoProc	getPrinterInfo;
initPrintProc		initPrint;
printBandProc		printBand;
endPrintProc		endPrint;

char const *matrixPath;

static int bidirectionalPrinting = FALSE;

/* returns the number of printer drivers, gets the location of dithering matrix */
int initPrintLib(const char *path)
{
#ifdef __MINGW32__
#ifdef DEBUG
 static char first = TRUE;

 /* so we can use printf in windows DLL */
 if (first)
  {
   AllocConsole();
   freopen("CONIN$", "r", stdin);
   freopen("CONOUT$", "w", stdout);
   freopen("CONOUT$", "w", stderr);
  }
 first = FALSE;
#endif
#endif
 matrixPath = path;
 return(NUM_PRINTERS);
}

/* sets bidirectional printing */
DREGISTER void setBidirectionalPrinting(int set)
{
 bidirectionalPrinting = set;
}

/* gets bidirectinal printing setting */
DREGISTER int getBidirectionalPrinting()
{
 return(bidirectionalPrinting);
}

/* return the default suffix for print file in the given printer format, used only for graphical */
/* formats, printers return NULL */
DREGISTER char *getPrintFileSuffix(int numPrinter)
{
 char *fileSuffix[] = {"png", "gif", "jpg", "pbm", "ps", "tif", "bmp" };
 
 return((numPrinter<7) ? fileSuffix[numPrinter]: NULL);
}

/* adds/replaces the suffix to the given filename. Warning! filename must have extra space for the length with */
/* the suffix! separator is used only if different than 0, suffix can have length 0 to get rid of .xxx */ 
DREGISTER void fixFilenameSuffix(char *fname, char separator, char *suffix)
{
 int flen, slen;

 if ((!suffix) || (!fname))
  return;
 flen = strlen(fname);
 slen = strlen(suffix);
 if (flen<(slen+2))
  {
   if (separator)
    { fname[flen]=separator; fname[flen+1]=0; }
  }
 else
  {
   if (separator)
    {
     if ((fname[flen-5]==separator) && ( !(isdigit((int) fname[flen-4]) && isdigit((int) fname[flen-3]) && isdigit((int) fname[flen-2]) && isdigit((int) fname[flen-1])) ) )
      fname[flen-4]=0;
     else
      if ((fname[flen-4]==separator) && ( !(isdigit((int) fname[flen-3]) && isdigit((int) fname[flen-2]) && isdigit((int) fname[flen-1])) ) )
       fname[flen-3]=0;
      else
       if ((fname[flen-3]==separator) && ( !(isdigit((int) fname[flen-2]) && isdigit((int) fname[flen-1])) ) )
        fname[flen-2]=0;
       else
        { fname[flen]=separator; fname[flen+1]=0; }
    }
   else
    if (!strcmp(suffix, &fname[flen-slen])) // equality test
     return;
  }
 if (slen)
  strcat(fname, suffix);
 else /* remove the separator dot, if we do not use the suffix */
  if (fname[strlen(fname)-1]=='.')
   fname[strlen(fname)-1]=0;
}

/* sets the functions of currently selected printer */
DREGISTER void setPrinterFunctions(int numPrinter)
{
 switch (numPrinter)
  {
   case 0: registerPNG(); break;
   case 1: registerGIF(); break;
   case 2: registerJPEG(); break;
   case 3: registerPbmPlus(); break;  
   case 4: registerPostScript(); break;  
   case 5: registerTIFF(); break;  
   case 6: registerBMP(); break;  
   case 7: registerStylus(); break;  
   case 8: registerStylusPro(); break;  
   case 9: registerStylusProXL(); break;  
   case 10: registerStylus800(); break;  
   case 11: registerStylus850(); break;  
   case 12: registerStylus1520(); break;  
   case 13: registerStylus3000(); break;  
   case 14: registerStylusPhoto(); break;  
   case 15: registerStylusPhoto700(); break; 
   case 16: registerStylusPhotoEX(); break;  
   case 17: registerStylus5000(); break;  
   case 18: registerStylusPhoto750(); break;  
   case 19: registerStylusPhoto1200(); break;  
   case 20: registerStylusPhoto870(); break;  
   case 21: registerStylusPhoto1270(); break;  
   case 22: registerStylusPhoto790(); break;  
   case 23: registerStylusPhoto890(); break;  
   case 24: registerStylusPhoto1290(); break;
#if !defined(__MINGW32__) && !defined(__APPLE__) 
   case 25: registerStylusPhotoEX3GutP(); break;  
   case 26: registerHPDeskJet850CGP(); break;  
   case 27: registerHPDeskJet1100CGP(); break;  
   case 28: registerHPDeskJet1120CGP(); break;  
   case 29: registerHPDeskJet1220CGP(); break;  
   case 30: registerHPCP1700CGP(); break;  

   case 31: registerPIXMAiP4000GutP(); break;  
   case 32: registerPIXMAiP4100GutP(); break;  
   case 33: registerPIXMAiP4200GutP(); break;  
   case 34: registerPIXMAiP4300GutP(); break;  
   case 35: registerPIXMAiP4500GutP(); break;  
   case 36: registerPIXMAiP5000GutP(); break;  
   case 37: registerPIXMAiP5200GutP(); break;  
   case 38: registerPIXMAiP5300GutP(); break;  
   case 39: registerPIXMAiP6000DGutP(); break;  
   case 40: registerPIXMAiP6700GutP(); break;  
   case 41: registerPIXMAiP8500GutP(); break;  
   case 42: registerPIXMAiX5000GutP(); break;  
   case 43: registerPIXMAPro9500GutP(); break;  
   case 44: registerPIXMAMP150GutP(); break;  
   case 45: registerPIXMAMP160GutP(); break;  
   case 46: registerPIXMAMP170GutP(); break;  
   case 47: registerPIXMAMP180GutP(); break;  
   case 48: registerPIXMAMP500GutP(); break;  
   case 49: registerPIXMAMP520GutP(); break;  
   case 50: registerPIXMAMP610GutP(); break;  
   case 51: registerPIXMAMP700GutP(); break;  
   case 52: registerPIXMAMP710GutP(); break;  
   case 53: registerPIXMAMP730GutP(); break;  
   case 54: registerPIXMAMP740GutP(); break;  
   case 55: registerPIXMAMP750GutP(); break;  
   case 56: registerPIXMAMP760GutP(); break;  
   case 57: registerPIXMAMP770GutP(); break;  
   case 58: registerPIXMAMP780GutP(); break;  
   case 59: registerPIXMAMP790GutP(); break;  
   case 60: registerPIXMAMP830GutP(); break;  

   case 61: registerStylusProXLGP(); break;  
   case 62: registerStylus3000GP(); break;  
   case 63: registerStylusPro5500GP(); break;  
   case 64: registerStylus9000GP(); break;  
   case 65: registerStylusC70GP(); break;  
   case 66: registerStylusC80GP(); break;  
   case 67: registerStylus880GP(); break;  
   case 68: registerStylusC64GP(); break;  
   case 69: registerStylusC82GP(); break;  
   case 70: registerStylusC88GutP(); break;  
   case 71: registerStylus670GP(); break;  
   case 72: registerStylus1160GP(); break;  
   case 73: registerStylusCX3200GP(); break;  
   case 74: registerStylusPhoto870GP(); break;  
   case 75: registerStylusPhoto1270GP(); break;  
   case 76: registerStylusPhoto890GP(); break;  
   case 77: registerStylusPhoto1290GP(); break;
   case 78: registerStylusPhoto830GP(); break;  
   case 79: registerStylusPhoto935GP(); break;
   case 80: registerStylusPhoto950GP(); break;
   case 81: registerStylusPhoto2100GP(); break;
   case 82: registerStylusPhoto2200GP(); break;
   case 83: registerStylusR200GP(); break;
   case 84: registerStylusR300GP(); break;
   case 85: registerStylusRX500GP(); break;
   case 86: registerStylusRX600GP(); break;

   case 87: registerStylusR200GutP(); break;
   case 88: registerStylusR230GutP(); break;
   case 89: registerStylusR270GutP(); break;
   case 90: registerStylusR285GutP(); break;
   case 91: registerStylusR290GutP(); break;
   case 92: registerStylusR350GutP(); break;
   case 93: registerStylusR360GutP(); break;
   case 94: registerStylusR390GutP(); break;
   case 95: registerStylusD68GutP(); break;
   case 96: registerStylusDX3850GutP(); break;
   case 97: registerStylusDX8400GutP(); break;
   case 98: registerStylusR800GutP(); break;
   case 99: registerStylusB300GutP(); break;
   case 100: registerStylusB500DNGutP(); break;
   case 101: registerStylusPhotoP50GutP(); break;
   case 102: registerStylusPhotoT60GutP(); break;
   
   case 103: registerStylus1400GutP(); break;
   case 104: registerStylusR1800GutP(); break;
   case 105: registerStylusR1900GutP(); break;
   case 106: registerStylusR2000GutP(); break;
   case 107: registerStylusR2400GutP(); break;
   case 108: registerStylusR3000GutP(); break;
   case 109: registerStylusD120GutP(); break;
   case 110: registerStylusME300GutP(); break;
   case 111: registerPS2A4GutP(); break;
   case 112: registerPS2A3GutP(); break;
#endif
  }
}


#define MATS  65536
/* allocates and returns the dither matrix of size 256*256 and distribution */
/* of values ranging from 0 to 255 according to the value of distr[256] */
/* array. The sum of all values in the distribution array must be 65536 */
DREGISTER unsigned char *makeDither(unsigned short *distr)
{
 unsigned char *dith;
 unsigned char mat[MATS*2], *matx, *maty;
 int i, distrPos;
 char filen[100];
 FILE *fp;

 matx=mat;
 maty=&mat[MATS];
 strcpy(filen, matrixPath);
 strcat(filen, "/matrix");
 if ((fp = fopen(filen, "rb" )) == NULL)
  { printf("open %s failed!\n", filen); return(NULL); }
 fread(mat, 1, MATS*2, fp);
 fclose(fp);

 dith = malloc(MATS);
 if (dith==NULL)
  return(NULL);

 distrPos=0;
 for (i=0; i<MATS; i++)
  {
   while ((distr[distrPos]==0)&&(distrPos<255)) distrPos++;
   dith[(maty[i]<<8) + matx[i]] = (unsigned char) distrPos;
   distr[distrPos]--;
  }
 return(dith);
}

/* registers printer with printer driver */
DREGISTER void registerPrinter(getPrinterInfoProc getPrtInfo,
 initPrintProc initPrt, printBandProc prtBand, endPrintProc endPrt)
{
 getPrinterInfo = getPrtInfo;
 initPrint = initPrt;
 printBand = prtBand;
 endPrint = endPrt;
}

/* return -1 if process can't be created, 0 on failure, 1 on success */
static int executeCommand(char *command)
{
 int status;

 status = system(command);
#ifdef __MINGW32__
    return status == 0 ? TRUE : FALSE;
#else
 if(status<0)
  return(-1);
 if (WIFEXITED(status))
  if (WEXITSTATUS(status)==0)
   return(TRUE);
  else
   return(FALSE);
 else
  return(FALSE);
#endif
}

static char *dstFile = NULL;
static char *commandStr = NULL;
static int doRemove = FALSE;

DREGISTER char *getRegistredDstFile(void)
{
 return(dstFile);
}

DREGISTER char *getRegistredCommandStr(void)
{
 return(commandStr);
}

DREGISTER int getRegistredDoRemove(void)
{
 return(doRemove);
}

/* set the parameters for the action at the end of printing: either */
/* - copy the temporary print file to dFile (using command comStr) */
/* - print the temporary print file (using command comStr) */
/* and optionally remove the source file and free the filename */
DREGISTER void setEnqueuePrintFile(char *dFile, char *comStr, int removeSource)
{
 if (dstFile)
  free(dstFile);
 dstFile = NULL;
 if (dFile)
  dstFile = strdup(dFile);

 if (commandStr)
  free(commandStr);
 commandStr = NULL;
 if (comStr)
  commandStr = strdup(comStr);
 doRemove = removeSource;
}

/* cleanup of temporary files */ 
DREGISTER void removeFileFreeFilename(char *filename, int prtStatus)
{
 if ((!dstFile) || (prtStatus!=PRT_OK))
  if (doRemove)
   if (access(filename, F_OK)==0)
    remove(filename);
 free(filename);
}

/* copy or move the file and set write permissions to the destination */
/* we used to fork a mv command, but this no longer works properly over CIFS networks */
/* anyway, we should be able to copy it on our own and we also avoid a fork */
DREGISTER int copyMoveFile(char *file1, char *file2, int move, int allowWriteToAll)
{
 FILE *inF, *outF;
 unsigned char buf[1<<16];
 int fileSize, i=0;
 int thisBlock;

//printf("move file %s to %s\n", file1, file2);
 if (access(file1, R_OK)!=0)
  return(FALSE);
 if ((inF = fopen(file1, "rb" )) == NULL)
  return(FALSE);
 if ((outF = fopen(file2, "wb" )) == NULL)
  {
   fclose(inF);
   return(FALSE);
  }

 fseek(inF, 0, SEEK_END);
 fileSize=ftell(inF);
 rewind(inF);
 /* buffered read/write, 65K at a time */
 while (i<fileSize)
  {
   thisBlock = min(1<<16, fileSize-i);
   fread(buf, 1, thisBlock, inF);
   fwrite(buf, 1, thisBlock, outF);
   i+=thisBlock;
  }

 fclose(inF);
 fclose(outF);
 if (allowWriteToAll)
  chmod(file2, 00666);
 if (move)
  remove(file1);
 return(TRUE);
}

/* does the proper thing (copy, move or print) with the finished print file */
DREGISTER int enqueuePrintFile(char *srcFile)
{
 char command[500];
 int status;

 /* give read/write permissions to everyone on source file */
 if (srcFile)
  if (chmod(srcFile, 00666)!=0)
   printf("aspawngp: Cannot change write permissions for %s\n", srcFile);

 if (!commandStr) /* need to move the file from srcFile to dstFile */
  {
   if (rename(srcFile, dstFile)==0)
    return(PRT_OK);
   else /* rename fails across filesystems, so move it by hand */
    {
     if (copyMoveFile(srcFile, dstFile, TRUE, TRUE))
      return(PRT_OK);
     else
      return(PRT_INTERNAL_ERROR);
    }
  }
 else
  strcpy(command, commandStr); /* use the provided command */

 /* take the command, add srcfile (and dstfile, if present) */
 strcat(command, " \"");
 strcat(command, srcFile);
 strcat(command, "\" ");
 if (dstFile)
  {
   strcat(command, " \"");
   strcat(command, dstFile);
   strcat(command, "\" ");
  }
 status = executeCommand(command);

 /* give read/write permissions to everyone on destination file */
 if (dstFile)
  if (chmod(dstFile, 00666)!=0)
   printf("save/print: Cannot change write permissions for %s\n", dstFile);

 if (status<0)
  return(PRT_INTERNAL_ERROR);
 else
  return((status==0) ? PRT_IO_ERROR : PRT_OK);
/* this code is unreachable, and apparently ArahPaint does not properly set doRemove, since it removes file after saving
 if (doRemove)
  if (access(dstFile, F_OK)==0)
   {
    if (doRemove)
     remove(dstFile);
    free(dstFile);
   }
*/
}

/* function takes bigger 8 bit image and reduces it to a smaller 24 bit image, resulting in antialiased view */
/* src must be 8bit image, dst must be 24 bit image */
/* source image must always be 4 times bigger in both dimensions */
DREGISTER void antialiasImage(unsigned char *srcImage, int srcBPL, unsigned char *dstImage, int dstBPL, unsigned char *cmap, int dstWid, int dstHei)
{
 int x, y, r, g, b, i;
 unsigned char *dstData, *srcData, *srcPix;
 unsigned int pixel;

 /* some paranoid checking */
 if ((!cmap) || (!srcImage) || (!dstImage))
  return;

 for (y=0; y<dstHei; y++)
  {
   srcData  = &srcImage[srcBPL*y<<2];
   dstData  = &dstImage[dstBPL*y];
   for (x=0; x<dstWid; x++)
    {
     r=g=b=0;

     srcPix = &srcData[(x<<2)];
     for (i=4; i; i--)
      {
       pixel = (*srcPix++)*3;
       r+=cmap[pixel++];
       g+=cmap[pixel++];
       b+=cmap[pixel];
       pixel = (*srcPix++)*3;
       r+=cmap[pixel++];
       g+=cmap[pixel++];
       b+=cmap[pixel];
       pixel = (*srcPix++)*3;
       r+=cmap[pixel++];
       g+=cmap[pixel++];
       b+=cmap[pixel];
       pixel = (*srcPix)*3;
       r+=cmap[pixel++];
       g+=cmap[pixel++];
       b+=cmap[pixel];
       srcPix += srcBPL-3;
      }

     *dstData++ = r>>4;
     *dstData++ = g>>4;
     *dstData++ = b>>4;
    }
  }
}

